#!/usr/bin/env python3
"""A backend toolkit for management of WebScrapBook collection.
"""
import sys
import os
from configparser import ConfigParser
from collections import OrderedDict
import mimetypes
import re

__all__ = ['WSB_USER_CONFIG', 'WSB_DIR', 'WSB_LOCAL_CONFIG', 'config']

__package_name__ = 'webscrapbook'
__version__ = '0.3.0'
__author__ = 'Danny Lin'
__author_email__ = 'danny0838@gmail.com'
__homepage__ = 'https://github.com/danny0838/PyWebScrapBook'
__license__ = 'MIT'

WSB_USER_CONFIG = os.path.join(os.path.expanduser('~'), '.wsbconfig')
WSB_DIR = '.wsb'
WSB_LOCAL_CONFIG = 'config.ini'

mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/html+zip", ".htz")
mimetypes.add_type("application/x-maff", ".maff")


class Config():
    """Config class whose values are lazily-initialized when accessed.

    Values are loaded from config files relative to CWD. Consider calling
    load() if the script changes CWD during the runtime.
    """
    def __init__(self):
        self._conf = None


    def __getitem__(self, key):
        if self._conf is None: self.load()  # lazy load
        return self._conf[key]


    @property
    def book(self):
        if self._conf is None: self.load()  # lazy load
        return self._book


    @property
    def auth(self):
        if self._conf is None: self.load()  # lazy load
        return self._auth


    def dump(self, fh):
        if self._conf is None: self.load()  # lazy load
        self._conf.write(fh)


    def dump_object(self):
        """Dump configs as an object, with type casting.
        """
        if self._conf is None: self.load()  # lazy load

        data = OrderedDict()
        for section in self._conf.sections():
            # remove [section "subsection"]
            if re.search(r'\s+"[^"]*"$', section):
                continue

            data[section] = OrderedDict()
            for key in self._conf[section]:
                data[section][key] = self._conf[section][key]

        # type casting
        data['server']['port'] = self._conf['server'].getint('port')
        data['server']['ssl_on'] = self._conf['server'].getboolean('ssl_on')
        data['auth'] = self._auth
        data['browser']['launch'] = self._conf['browser'].getboolean('launch')
        data['browser']['new'] = self._conf['browser'].getint('new')
        data['browser']['top'] = self._conf['browser'].getboolean('top')
        data['book'] = self._book

        return data


    def load(self, root='.'):
        """Loads config files related to the given root directory.

        Skip if config file doesn't exist, but raise if not loadable.

        project config > user config > default config
        """
        def load_config(file):
            if not os.path.isfile(file):
                return

            try:
                parser = ConfigParser(interpolation=None)
                parser.read(file, encoding='UTF-8')
                for section in parser.sections():
                    if section not in conf:
                        conf[section] = {}
                    for key in parser[section]:
                        conf[section][key] = parser[section][key]
            except:
                print('Error: Unable to load config from "{}".'.format(file), file=sys.stderr)
                raise


        def parse_subsection(name):
            """Hanble subsection formatted as [section "subsection"].

            Also map [section] to [section ""].
            """
            data = OrderedDict()
            for section in conf:
                m = re.search(r'^' + name + r'(?:\s*"([^"\]]*)")?$', section)
                if not m: continue
                c = data.setdefault(m.group(1) or '', OrderedDict())
                for key in conf[section]:
                    c[key] = conf[section][key]
            return data


        # default config
        self._conf = conf = ConfigParser(interpolation=None)
        conf['app'] = {}
        conf['app']['name'] = 'WebScrapBook'
        conf['app']['theme'] = 'default'
        conf['app']['root'] = '.'
        conf['app']['base'] = ''
        conf['server'] = {}
        conf['server']['port'] = '8080'
        conf['server']['bind'] = '127.0.0.1'
        conf['server']['host'] = 'localhost'
        conf['server']['threads'] = '0'
        conf['server']['ssl_on'] = 'false'
        conf['server']['ssl_key'] = ''
        conf['server']['ssl_cert'] = ''
        conf['server']['ssl_pw'] = ''
        conf['server']['browse'] = 'true'
        conf['browser'] = {}
        conf['browser']['command'] = ''
        conf['browser']['new'] = '0'
        conf['browser']['top'] = 'false'
        conf['browser']['index'] = ''
        conf['book'] = {}
        conf['book']['name'] = 'scrapbook'
        conf['book']['top_dir'] = ''
        conf['book']['data_dir'] = ''
        conf['book']['tree_dir'] = '.wsb/tree'
        conf['book']['index'] = '.wsb/tree/map.html'

        # user config
        load_config(WSB_USER_CONFIG)

        # book config
        load_config(os.path.join(root, WSB_DIR, WSB_LOCAL_CONFIG))

        # handle subsections
        self._book = parse_subsection('book')
        self._auth = parse_subsection('auth')

config = Config()



#!/usr/bin/env python3
"""The WGSI application.

   Ways to host this app:
   1. run server.serve()
   2. set application = get_app() in a .wsgi
"""
import sys
import os
import traceback
import shutil
import mimetypes
import re
import zipfile
import time
import email.utils
import hashlib
import json
from base64 import b64decode
from urllib.parse import urlsplit, urlunsplit, urljoin, quote, unquote, parse_qs

# dependency
import bottle
from bottle import request, redirect, static_file, template
from bottle import HTTPResponse, HTTPError
import commonmark

# this package
from . import *
from . import util

# runtime variables
# initializated in get_app() just before app is started
runtime = {}

# main app instance
app = bottle.Bottle()

# patch a bug in Bottle 0.12.*
# https://github.com/bottlepy/bottle/issues/1129
def patch_bottle_0_12():
    from distutils.version import LooseVersion

    if LooseVersion(bottle.__version__) >= LooseVersion("0.13"):
        return

    from bottle import parse_range_header, parse_date, _file_iter_range, tob

    global static_file

    # implement static_file() in 0.13
    # removed "close" param of _file_iter_range as it's not implemented in 0.12 yet.
    def static_file(filename, root,
                    mimetype=True,
                    download=False,
                    charset='UTF-8',
                    etag=None):
        """ Open a file in a safe way and return an instance of :exc:`HTTPResponse`
            that can be sent back to the client.
            :param filename: Name or path of the file to send, relative to ``root``.
            :param root: Root path for file lookups. Should be an absolute directory
                path.
            :param mimetype: Provide the content-type header (default: guess from
                file extension)
            :param download: If True, ask the browser to open a `Save as...` dialog
                instead of opening the file with the associated program. You can
                specify a custom filename as a string. If not specified, the
                original filename is used (default: False).
            :param charset: The charset for files with a ``text/*`` mime-type.
                (default: UTF-8)
            :param etag: Provide a pre-computed ETag header. If set to ``False``,
                ETag handling is disabled. (default: auto-generate ETag header)
            While checking user input is always a good idea, this function provides
            additional protection against malicious ``filename`` parameters from
            breaking out of the ``root`` directory and leaking sensitive information
            to an attacker.
            Read-protected files or files outside of the ``root`` directory are
            answered with ``403 Access Denied``. Missing files result in a
            ``404 Not Found`` response. Conditional requests (``If-Modified-Since``,
            ``If-None-Match``) are answered with ``304 Not Modified`` whenever
            possible. ``HEAD`` and ``Range`` requests (used by download managers to
            check or continue partial downloads) are also handled automatically.
        """

        root = os.path.join(os.path.abspath(root), '')
        filename = os.path.abspath(os.path.join(root, filename.strip('/\\')))
        headers = dict()

        if not filename.startswith(root):
            return HTTPError(403, "Access denied.")
        if not os.path.exists(filename) or not os.path.isfile(filename):
            return HTTPError(404, "File does not exist.")
        if not os.access(filename, os.R_OK):
            return HTTPError(403, "You do not have permission to access this file.")

        if mimetype is True:
            if download and download is not True:
                mimetype, encoding = mimetypes.guess_type(download)
            else:
                mimetype, encoding = mimetypes.guess_type(filename)
            if encoding: headers['Content-Encoding'] = encoding

        if mimetype:
            if (mimetype[:5] == 'text/' or mimetype == 'application/javascript')\
            and charset and 'charset' not in mimetype:
                mimetype += '; charset=%s' % charset
            headers['Content-Type'] = mimetype

        if download:
            download = os.path.basename(filename if download is True else download)
            headers['Content-Disposition'] = 'attachment; filename="%s"' % download

        stats = os.stat(filename)
        headers['Content-Length'] = clen = stats.st_size
        headers['Last-Modified'] = email.utils.formatdate(stats.st_mtime,
                                                          usegmt=True)
        headers['Date'] = email.utils.formatdate(time.time(), usegmt=True)

        getenv = request.environ.get

        if etag is None:
            etag = '%d:%d:%d:%d:%s' % (stats.st_dev, stats.st_ino, stats.st_mtime,
                                       clen, filename)
            etag = hashlib.sha1(tob(etag)).hexdigest()

        if etag:
            headers['ETag'] = etag
            check = getenv('HTTP_IF_NONE_MATCH')
            if check and check == etag:
                return HTTPResponse(status=304, **headers)

        if not (etag and check):
            ims = getenv('HTTP_IF_MODIFIED_SINCE')
            if ims:
                ims = parse_date(ims.split(";")[0].strip())
            if ims is not None and ims >= int(stats.st_mtime):
                return HTTPResponse(status=304, **headers)

        body = '' if request.method == 'HEAD' else open(filename, 'rb')

        headers["Accept-Ranges"] = "bytes"
        range_header = getenv('HTTP_RANGE')
        if range_header:
            ranges = list(parse_range_header(range_header, clen))
            if not ranges:
                return HTTPError(416, "Requested Range Not Satisfiable")
            offset, end = ranges[0]
            headers["Content-Range"] = "bytes %d-%d/%d" % (offset, end - 1, clen)
            headers["Content-Length"] = str(end - offset)
            if body: body = _file_iter_range(body, offset, end - offset)
            return HTTPResponse(body, status=206, **headers)
        return HTTPResponse(body, **headers)

    bottle.static_file = static_file

patch_bottle_0_12()


def http_response(body='', status=None, headers=None, format=None, **more_headers):
    """Handles formatted response.

    ref: https://jsonapi.org
    """
    if format == 'json':
        more_headers['Content-type'] = 'application/json'

        body = {
            'success': True,
            'data': body,
            }

        body = json.dumps(body, ensure_ascii=False)

    return HTTPResponse(body, status, headers, **more_headers)


def http_error(
         status=500,
         body=None,
         exception=None,
         traceback=None,
         format=None, **more_headers):
    """Handles formatted error response.
    """
    if format == 'json':
        more_headers['Content-type'] = 'application/json'

        body = {
            'error': {
                'status': status,
                'message': body,
                },
            }
        body = json.dumps(body, ensure_ascii=False)
        return HTTPResponse(body, status, **more_headers)

    else:
        return HTTPError(status, body, exception, traceback, **more_headers)


def get_base():
    return (config['app']['base'] or request.environ.get('SCRIPT_NAME', '')).rstrip('/')


def get_pathname():
    """Revise request.pathname of Bottle

    When visiting http://example.com/app, request.pathname returns '/app/',
    while we want '/app'.

    http://example.com effectively means http://example.com/, so keep at least
    '/' as final fallback.
    """
    return (request.environ.get('SCRIPT_NAME', '') + request.environ.get('PATH_INFO', '')) or '/'


def get_archive_path(filepath, localpath):
    """Parse archive file path and the sub-archive path.

    - Priority:
      /path/to/fileordir.htz!
      /path/to/fileordir.htz
      /path/to/fileordir.htz!/path/to/fileordir.htz!
      /path/to/fileordir.htz!/path/to/fileordir.htz
      ...

    - Returns a tuple (archivefile, subarchivepath).
    """
    if not os.path.lexists(localpath):
        for m in re.finditer(r'![/\\]', filepath, flags=re.I):
            archivefile = os.path.join(runtime['root'], filepath[:m.start(0)].strip('/\\'))
            if os.path.isdir(archivefile + '!'):
                return (None, None)

            if zipfile.is_zipfile(archivefile):
                subarchivepath = filepath[m.end(0):].rstrip('/')
                return (archivefile, subarchivepath)

    return (None, None)


def is_local_access():
    """Determine if the client is in same device.
    """
    return request.urlparts.hostname in ('localhost', '127.0.0.1', request.remote_addr)


def handle_authorization(format=None):
    """Check if authorized or not.

    Return None if authorization passed, otherwise the header and body for authorization.

    Also modify runtime['permission'] for further access control.
    """
    if not len(config.auth):
        return

    auth_pass = False
    user, pw = request.auth or ('', '')

    # The higher level (e.g. Apache if this is run via Apache WSGI) may remove
    # the password and gets None. Revise to '' to prevent further error.
    if pw is None: pw = ''

    if user:
        for _, entry in config.auth.items():
            entry_user = entry.get('user', '')
            entry_pw = entry.get('pw', '')
            entry_pw_salt = entry.get('pw_salt', '')
            entry_pw_type = entry.get('pw_type', '')
            entry_permission = entry.get('permission', 'all')
            if (user == entry_user and
                    util.encrypt(pw, entry_pw_salt, entry_pw_type) == entry_pw):
                auth_pass = True
                runtime['permission'] = entry.get('permission', '')
                break

    if not auth_pass:
        headers = {
            'WWW-Authenticate': 'Basic realm="Authentication required.", charset="UTF-8"',
            'Content-type': 'text/html',
            }
        return http_error(401, "You are not authorized.", format=format, **headers)


def handle_directory_listing(localpath, format=None):
    """List contents in a directory.
    """
    # ensure directory has trailing '/'
    pathname = get_pathname()
    if not pathname.endswith('/'):
        parts = request.urlparts
        new_parts = (parts[0], parts[1], quote(pathname) + '/', parts[3], parts[4])
        new_url = urlunsplit(new_parts)
        return redirect(new_url)

    if not os.path.exists(localpath):
        return http_error(404, "Directory does not exist.", format=format)

    if not os.access(localpath, os.R_OK):
        return http_error(403, "You do not have permission to view this directory.", format=format)

    # headers
    headers = {}
    stats = os.stat(localpath)
    headers['Last-Modified'] = email.utils.formatdate(stats.st_mtime, usegmt=True)
    headers['Date'] = email.utils.formatdate(time.time(), usegmt=True)
    headers['Cache-Control'] = 'no-store'

    # output index
    subentries = util.listdir(localpath)

    if format:
        data = []
        for entry in subentries:
            data.append({
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    })
        return http_response(data, format=format, **headers)

    body = template('index.tpl',
            sitename=runtime['name'],
            is_local=is_local_access(),
            base=get_base(),
            path=request.path,
            subarchivepath=None,
            subentries=subentries,
            )

    return HTTPResponse(body, **headers)


def handle_zip_directory_listing(zip, archivefile, subarchivepath, format=None):
    """List contents in a directory.
    """
    # ensure directory has trailing '/'
    pathname = get_pathname()
    if not pathname.endswith('/'):
        parts = request.urlparts
        new_parts = (parts[0], parts[1], quote(pathname) + '/', parts[3], parts[4])
        new_url = urlunsplit(new_parts)
        return redirect(new_url)

    from bottle import parse_date, tob

    headers = {}

    # check 304
    stats = os.stat(archivefile)
    headers['Last-Modified'] = email.utils.formatdate(stats.st_mtime, usegmt=True)
    headers['Cache-Control'] = 'no-cache'

    getenv = request.environ.get

    etag = '%d:%d:%d:%d:%s' % (stats.st_dev, stats.st_ino, stats.st_mtime,
                               stats.st_size, archivefile)
    etag = hashlib.sha1(tob(etag)).hexdigest()

    headers['ETag'] = etag
    check = getenv('HTTP_IF_NONE_MATCH')
    if check and check == etag:
        return HTTPResponse(status=304, **headers)

    if not check:
        ims = getenv('HTTP_IF_MODIFIED_SINCE')
        if ims:
            ims = parse_date(ims.split(";")[0].strip())
        if ims is not None and ims >= int(stats.st_mtime):
            return HTTPResponse(status=304, **headers)

    subentries = util.zip_listdir(zip, subarchivepath)

    try:
        body = template('index.tpl',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=get_base(),
                path=request.path,
                subarchivepath=subarchivepath,
                subentries=subentries,
                )
        return HTTPResponse(body, **headers)
    except util.ZipDirNotFoundError:
        return http_error(404, "File does not exist.", format=format)


def handle_subarchive_path(
        archivefile,
        subarchivepath,
        mimetype,
        encoding,
        download=False,
        charset='UTF-8',
        etag=None,
        format=None):
    """Show content of a path in a zip file.
    """
    from bottle import parse_range_header, parse_date, _file_iter_range, tob

    if not os.access(archivefile, os.R_OK):
        return http_error(403, "You do not have permission to access this file.", format=format)

    try:
        zip = zipfile.ZipFile(archivefile)
    except:
        return http_error(500, "Unable to open the ZIP file.", format=format)

    try:
        # KeyError is raised if subarchivepath does not exist
        info = zip.getinfo(subarchivepath)
    except KeyError:
        # subarchivepath does not exist
        # possibility a missing directory entry?
        return handle_zip_directory_listing(zip, archivefile, subarchivepath)

    fh = zip.open(subarchivepath, 'r')

    headers = dict()

    if encoding: headers['Content-Encoding'] = encoding

    if mimetype is True:
        if download and download is not True:
            mimetype, encoding = mimetypes.guess_type(download)
        else:
            mimetype, encoding = mimetypes.guess_type(subarchivepath)
        if encoding: headers['Content-Encoding'] = encoding

    if mimetype:
        if (mimetype[:5] == 'text/' or mimetype == 'application/javascript')\
        and charset and 'charset' not in mimetype:
            mimetype += '; charset=%s' % charset
        headers['Content-Type'] = mimetype

    if download:
        download = os.path.basename(subarchivepath if download is True else download)
        headers['Content-Disposition'] = 'attachment; filename="%s"' % download

    headers['Content-Length'] = clen = info.file_size

    lm = info.date_time
    epoch = int(time.mktime((lm[0], lm[1], lm[2], lm[3], lm[4], lm[5], 0, 0, -1)))
    headers['Last-Modified'] = email.utils.formatdate(epoch, usegmt=True)

    headers['Date'] = email.utils.formatdate(time.time(), usegmt=True)

    getenv = request.environ.get

    if etag is None:
        etag = '%d:%d:%s' % (epoch, clen, subarchivepath)
        etag = hashlib.sha1(tob(etag)).hexdigest()

    if etag:
        headers['ETag'] = etag
        check = getenv('HTTP_IF_NONE_MATCH')
        if check and check == etag:
            return HTTPResponse(status=304, **headers)

    if not (etag and check):
        ims = getenv('HTTP_IF_MODIFIED_SINCE')
        if ims:
            ims = parse_date(ims.split(";")[0].strip())
        if ims is not None and ims >= int(epoch):
            return HTTPResponse(status=304, **headers)

    body = '' if request.method == 'HEAD' else fh

    headers["Accept-Ranges"] = "bytes"
    range_header = getenv('HTTP_RANGE')
    if range_header:
        ranges = list(parse_range_header(range_header, clen))
        if not ranges:
            return http_error(416, "Requested Range Not Satisfiable")
        offset, end = ranges[0]
        headers["Content-Range"] = "bytes %d-%d/%d" % (offset, end - 1, clen)
        headers["Content-Length"] = str(end - offset)
        if body: body = _file_iter_range(body, offset, end - offset)
        return HTTPResponse(body, status=206, **headers)
    return HTTPResponse(body, **headers)


def handle_archive_viewing(localpath, mimetype):
    """Handle direct visit of HTZ/MAFF file.
    """
    def list_maff_pages(pages):
        """List available web pages in a MAFF file.
        """
        return template('maff_index.tpl',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=get_base(),
                path=request.path,
                pages=pages,
                )

    if not os.access(localpath, os.R_OK):
        return http_error(403, "You do not have permission to access this file.")

    if mimetype == "application/html+zip":
        subpath = "index.html"
    else:
        pages = util.get_maff_pages(localpath)

        if len(pages) > 1:
            # multiple index files
            return list_maff_pages(pages)
        elif len(pages) > 0:
            subpath = pages[0].indexfilename
        else:
            # no valid index file found
            return list_maff_pages([])

    parts = request.urlparts
    new_parts = (parts[0], parts[1], parts[2] + '!/' + quote(subpath), parts[3], parts[4])
    new_url = urlunsplit(new_parts)
    return redirect(new_url)


def handle_markdown_output(filepath, filename):
    """Output processed markdown.
    """
    from bottle import parse_date, tob

    headers = {}

    # check 304
    stats = os.stat(filename)
    headers['Last-Modified'] = email.utils.formatdate(stats.st_mtime, usegmt=True)

    getenv = request.environ.get

    etag = '%d:%d:%d:%d:%s' % (stats.st_dev, stats.st_ino, stats.st_mtime,
                               stats.st_size, filename)
    etag = hashlib.sha1(tob(etag)).hexdigest()

    headers['ETag'] = etag
    check = getenv('HTTP_IF_NONE_MATCH')
    if check and check == etag:
        return HTTPResponse(status=304, **headers)

    if not check:
        ims = getenv('HTTP_IF_MODIFIED_SINCE')
        if ims:
            ims = parse_date(ims.split(";")[0].strip())
        if ims is not None and ims >= int(stats.st_mtime):
            return HTTPResponse(status=304, **headers)

    # output processed content
    with open(filename, 'r', encoding='UTF-8') as f:
        body = f.read()
        f.close()
    
    body = template('markdown',
            sitename=runtime['name'],
            is_local=is_local_access(),
            base=get_base(),
            path=request.path,
            content=commonmark.commonmark(body),
            )

    return HTTPResponse(body, **headers)


@app.route('<filepath:path>', method=['GET', 'HEAD', 'POST'])
def handle_request(filepath):
    """Handle an HTTP request (HEAD, GET, POST).
    """
    action = request.params.getunicode('a', encoding='UTF-8', default='view')
    action = request.params.getunicode('action', encoding='UTF-8', default=action)

    format = request.params.getunicode('f', encoding='UTF-8')
    format = request.params.getunicode('format', encoding='UTF-8', default=format)

    # handle authentication
    runtime['permission'] = 'all'
    auth_result = handle_authorization(format=format)
    if auth_result is not None:
        return auth_result

    # determine primary variables
    #
    # filepath: the URL path below app base (not percent encoded)
    # localpath: the file system path corresponding to filepath
    # localtargetpath: localpath with symbolic link resolved
    # mimetype: the mimetype from localtargetpath
    # archivefile: the file system path of the ZIP archive file, or None
    # subarchivepath: the URL path below archivefile (not percent encoded)
    localpath = os.path.abspath(os.path.join(runtime['root'], filepath.strip('/\\')))
    localtargetpath = os.path.realpath(localpath)
    archivefile, subarchivepath = get_archive_path(filepath, localpath)
    mimetype, encoding = mimetypes.guess_type(localtargetpath)

    # handle action
    if action == 'static':
        return static_file(filepath, root=runtime['static'])

    elif action == 'source':
        # show text-like files as plain text
        if mimetype and (
                mimetype.startswith('text/') or
                mimetype.endswith('+xml') or
                mimetype.endswith('+json') or
                mimetype in ('application/javascript')
                ):
            mimetype = 'text/plain; charset=UTF-8'

        return static_file(filepath, root=runtime['root'], mimetype=mimetype)

    elif action in ('exec', 'browse'):
        if is_local_access():
            if os.path.lexists(localpath):
                if action == 'browse':
                    util.view_in_explorer(localpath)

                elif action == 'exec':
                    util.launch(localpath)

                if format:
                    return http_response('Command run successfully.', format=format)
                else:
                    return http_response(status=204, format=format)
       
            return http_error(404, "File does not exist.", format=format)

        else:
            return http_error(400, "Command can only run on local device.", format=format)

    elif action == 'token':
        if runtime['permission'] != 'all':
            return http_error(401, "You are not permitted to do this.", format=format)

        return http_response(token_handler.acquire(), format=format)

    elif action == 'list':
        if not format:
            return http_error(400, "Action not supported.", format=format)

        if os.path.isdir(localpath):
            return handle_directory_listing(localtargetpath, format=format)

        return http_error(400, "This is not a directory.", format=format)

    elif action == 'config':
        if not format:
            return http_error(400, "Action not supported.", format=format)

        # slightly adjusted some value for client to better know the server
        data = config.dump_object()
        data['app']['is_local'] = is_local_access()
        data['app']['root'] = runtime['root']
        data['app']['base'] = get_base()
        return http_response(data, format=format)

    elif action == 'edit':
        if format:
            return http_error(400, "Action not supported.", format=format)

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            return http_error(400, "Found a non-file here.", format=format)

        try:
            with open(localpath, 'rb') as f:
                body = f.read()
                f.close()
        except FileNotFoundError:
            body = b''

        try:
            body = body.decode('UTF-8')
            is_utf8 = True
        except UnicodeDecodeError:
            body = body.decode('ISO-8859-1')
            is_utf8 = False

        body = template('edit.tpl',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=get_base(),
                path=request.path,
                body=body,
                is_utf8=is_utf8,
                )

        return http_response(body, format=format)

    elif action == 'editx':
        if format:
            return http_error(400, "Action not supported.", format=format)

        if not os.path.lexists(localpath):
            return http_error(404, "File does not exist.", format=format)

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            return http_error(400, "Found a non-file here.", format=format)

        if not mimetype in ("text/html", "application/xhtml+xml"):
            return http_error(400, "This is not an HTML file.", format=format)

        body = template('editx.tpl',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=get_base(),
                path=request.path,
                )

        return http_response(body, format=format)

    elif action in ('mkdir', 'save', 'upload', 'delete', 'move', 'copy'):
        if request.method != 'POST':
            headers = {
                'Allow': 'POST',
                }
            return http_error(405, 'Method "{}" not allowed.'.format(request.method), format=format, **headers)

        token = request.params.get('token') or ''

        if not token_handler.validate(token):
            return http_error(401, 'Invalid access token.', format=format)

        token_handler.delete(token)

        if os.path.abspath(localpath) == runtime['root']:
            return http_error(403, "Unable to operate the root directory.", format=format)

        targetpath = target = request.forms.get('target')

        if target is not None:
            targetpath = os.path.normpath(os.path.join(runtime['root'], target.strip('/')))
            if not targetpath.startswith(os.path.join(runtime['root'], '')):
                return http_error(403, "Unable to operate beyond the root directory.", format=format)

        if action == 'mkdir':
            if os.path.lexists(localpath) and not os.path.isdir(localpath):
                return http_error(400, "Found a non-directory here.", format=format)

            try:
                os.makedirs(localpath, exist_ok=True)
            except OSError:
                traceback.print_exc()
                return http_error(500, "Unable to create a directory here.", format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            return redirect(request.url)

        elif action == 'save':
            if os.path.lexists(localpath) and not os.path.isfile(localpath):
                return http_error(400, "Found a non-file here.", format=format)

            try:
                os.makedirs(os.path.dirname(localpath), exist_ok=True)
                bytes = request.forms.get('text').encode('ISO-8859-1') or b''
                with open(localpath, 'wb') as f:
                    f.write(bytes)
                    f.close()
            except OSError:
                traceback.print_exc()
                return http_error(500, "Unable to write to this file.", format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            return redirect(request.url)

        elif action == 'upload':
            if os.path.lexists(localpath):
                if not os.path.isfile(localpath):
                    return http_error(400, "Found a non-file here.", format=format)
                else:
                    os.remove(localpath)

            try:
                os.makedirs(os.path.dirname(localpath), exist_ok=True)
                file = request.files.get('upload')
                file.save(localpath)
            except OSError:
                traceback.print_exc()
                return http_error(500, "Unable to upload this file.", format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            return redirect(request.url)

        elif action == 'delete':
            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

            if os.path.islink(localpath):
                try:
                    os.remove(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this link.", format=format)
            elif os.path.isfile(localpath):
                try:
                    os.remove(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this file.", format=format)
            elif os.path.isdir(localpath):
                try:
                    shutil.rmtree(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this directory.", format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            parts = request.urlparts
            new_parts = (parts[0], parts[1], re.sub(r'[^/]+/?$', r'', parts[2]), '', '')
            new_url = urlunsplit(new_parts)
            return redirect(new_url)

        elif action == 'move':
            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

            if os.path.lexists(targetpath):
                return http_error(400, 'Found something at target "{}".'.format(target), format=format)
            else:
                os.makedirs(os.path.dirname(targetpath), exist_ok=True)

            try:
                os.rename(localpath, targetpath)
            except:
                traceback.print_exc()
                return http_error(500, 'Unable to move to target "{}".'.format(target), format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            parts = request.urlparts
            new_parts = (parts[0], parts[1], re.sub(r'[^/]+/?$', r'', parts[2]), '', '')
            new_url = urlunsplit(new_parts)
            return redirect(new_url)

        elif action == 'copy':
            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

            if os.path.lexists(targetpath):
                return http_error(400, 'Found something at target "{}".'.format(target), format=format)
            else:
                os.makedirs(os.path.dirname(targetpath), exist_ok=True)

            try:
                shutil.copytree(localpath, targetpath)
            except:
                try:
                    shutil.copy2(localpath, targetpath)
                except:
                    traceback.print_exc()
                    return http_error(500, 'Unable to copy to target "{}".'.format(target), format=format)

            if format:
                return http_response('Command run successfully.', format=format)

            return redirect(new_url)

    # "view" or unknown actions
    else:
        # show file information for other output formats
        if format:
            info = util.file_info(localpath)
            data = {
                'name': info.name,
                'type': info.type,
                'size': info.size,
                'last_modified': info.last_modified,
                'mime': mimetype,
                }
            return http_response(data, format=format)

        # handle directory
        if os.path.isdir(localpath):
            return handle_directory_listing(localtargetpath, format=format)

        # handle file
        elif os.path.isfile(localpath):
            # view archive file
            if mimetype in ("application/html+zip", "application/x-maff"):
                return handle_archive_viewing(localtargetpath, mimetype)

            # view markdown
            if mimetype == "text/markdown":
                return handle_markdown_output(filepath, localtargetpath)

            # convert meta refresh to 302 redirect
            if localtargetpath.lower().endswith('.htm'):
                target = util.parse_meta_refresh(localtargetpath).target

                if target is not None:
                    # Keep several chars as javascript encodeURI do,
                    # plus "!" to keep "!/" redirect work.
                    new_url = urljoin(request.url, quote(target, ',/?:@&=+$#!'))
                    return redirect(new_url)

            # show static file for other cases
            return static_file(filepath, root=runtime['root'], mimetype=mimetype)

        # handle sub-archive path
        elif archivefile:
            return handle_subarchive_path(os.path.realpath(archivefile), subarchivepath,
                    mimetype, encoding, format=format)

        # probably 404 not found here
        return static_file(filepath, root=runtime['root'], mimetype=mimetype)


def debug(*msg):
    """Quick dirty logging for testing purpose.
    """
    os.makedirs(os.path.dirname(runtime['log']), exist_ok=True)
    with open(runtime['log'], 'a', encoding='UTF-8') as f:
        f.write(' '.join(str(s) for s in msg))
        f.write('\n')
        f.close()


def init_app():
    """Run initialization and return the app.
    """
    runtime['root'] = os.path.abspath(config['app']['root'])
    runtime['name'] = config['app']['name']

    # if a local theme is found, overwrite the default theme
    runtime['theme'] = os.path.join(runtime['root'], WSB_DIR, 'themes', config['app']['theme'])
    if not os.path.isdir(runtime['theme']):
        runtime['theme'] = os.path.join(os.path.dirname(__file__), 'themes', config['app']['theme'])

    runtime['static'] = os.path.join(runtime['theme'], 'static')
    runtime['templates'] = os.path.join(runtime['theme'], 'templates')
    bottle.TEMPLATE_PATH.insert(0, runtime['templates'])

    # init token_handler
    global token_handler
    token_handler = util.TokenHandler(os.path.join(runtime['root'], WSB_DIR, 'server', 'token'))

    # init debugging logger
    runtime['log'] = os.path.join(runtime['root'], WSB_DIR, 'server', 'logs', 'debug.log')

    return app



#!/usr/bin/env python3
"""Command line interface of WebScrapBook toolkit.
"""
import sys
import os
import argparse
import json
from getpass import getpass

# this package
from . import __package_name__, __version__
from . import *
from . import server
from . import util


class WebScrapBookInitError(Exception):
    pass


def fcopy(fsrc, fdst):
    """Copy a script file to target

    - Use universal linefeed.
    - Set last modified time to current time.
    """
    os.makedirs(os.path.dirname(os.path.abspath(fdst)), exist_ok=True)
    with open(fsrc, 'r', encoding='UTF-8') as f:
        content = f.read()
        f.close()
    with open(fdst, 'w', encoding='UTF-8') as f:
        f.write(content)
        f.close()


def cmd_serve(args):
    """Serve the directory."""
    server.serve(args['root'])


def cmd_config(args):
    """Show or edit config."""
    if args['book']:
        filename = WSB_LOCAL_CONFIG
        fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, filename))
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', filename))
        if not os.path.isfile(fdst):
            try:
                print('Generating "{}"...'.format(fdst))
                fcopy(fsrc, fdst)
            except:
                raise WebScrapBookInitError("Unable to generate {}.".format(fdst))

        try:
            util.launch(fdst)
        except OSError:
            pass

        if args['all']:
            filename = 'serve.py'
            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, filename))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', filename))
            if not os.path.isfile(fdst):
                try:
                    print('Generating "{}"...'.format(fdst))
                    fcopy(fsrc, fdst)
                except:
                    raise WebScrapBookInitError("Unable to generate {}.".format(fdst))

            filename = 'serve.wsgi'
            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, filename))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', filename))
            if not os.path.isfile(fdst):
                try:
                    print('Generating "{}"...'.format(fdst))
                    fcopy(fsrc, fdst)
                except:
                    raise WebScrapBookInitError("Unable to generate {}.".format(fdst))

    if args['user']:
        fdst = WSB_USER_CONFIG
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', WSB_LOCAL_CONFIG))
        if not os.path.isfile(fdst):
            try:
                print('Generating "{}"...'.format(fdst))
                fcopy(fsrc, fdst)
            except:
                raise WebScrapBookInitError("Unable to generate {}.".format(fdst))

        try:
            util.launch(fdst)
        except OSError:
            pass

    if not any(args[k] for k in ('book', 'user')):
        config.dump(sys.stdout)


def cmd_encrypt(args):
    """Generate encrypted password string."""
    if args['password'] is None:
        pw1 = getpass('Enter a password: ')
        pw2 = getpass('Confirm the password: ')

        if pw1 != pw2:
            print('Error: Entered passwords do not match.', file=sys.stderr)
            sys.exit(1)

        args['password'] = pw1

    print(util.encrypt(args['password'], salt=args['salt'], method=args['method']))


def cmd_help(args):
    """Show detailed information."""
    root = os.path.join(os.path.dirname(__file__), 'resources')

    if args['topic'] == 'config':
        file = os.path.join(root, 'config.md')
        with open(file, 'r', encoding='UTF-8') as f:
            text = f.read()
            f.close()
        print(text)


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--version', default=False, action='store_true',
        help="""show version information and exit""")
    parser.add_argument('--root', default=".",
        help="""root directory to manipulate (default: current working directory)""")
    subparsers = parser.add_subparsers(dest='command', metavar='COMMAND')

    # subcommand: serve
    parser_serve = subparsers.add_parser('serve', aliases=['s'],
        help=cmd_serve.__doc__, description=cmd_serve.__doc__)
    parser_serve.set_defaults(func=cmd_serve)

    # subcommand: config
    parser_config = subparsers.add_parser('config', aliases=['c'],
        help=cmd_config.__doc__, description=cmd_config.__doc__)
    parser_config.set_defaults(func=cmd_config)
    parser_config.add_argument('-b', '--book', default=False, action='store_true',
        help="""generate and edit book config.""")
    parser_config.add_argument('-u', '--user', default=False, action='store_true',
        help="""generate and edit user config.""")
    parser_config.add_argument('-a', '--all', default=False, action='store_true',
        help="""generate more assistent files.""")

    # subcommand: encrypt
    parser_encrypt = subparsers.add_parser('encrypt', aliases=['e'],
        help=cmd_encrypt.__doc__, description=cmd_encrypt.__doc__)
    parser_encrypt.set_defaults(func=cmd_encrypt)
    parser_encrypt.add_argument('-p', '--password', nargs='?', default=None, action='store',
        help="""the password to encrypt.""")
    parser_encrypt.add_argument('-m', '--method', default='sha1', action='store',
        help="""the encrypt method to use, which is one of: plain, md5, sha1,
sha224, sha256, sha384, sha512, sha3_224, sha3_256, sha3_384, and sha3_512.
(default: %(default)s)""")
    parser_encrypt.add_argument('-s', '--salt', default='', action='store',
        help="""the salt to add during encryption.""")

    # subcommand: help
    parser_help = subparsers.add_parser('help',
        help=cmd_help.__doc__, description=cmd_help.__doc__)
    parser_help.set_defaults(func=cmd_help)
    parser_help.add_argument('topic', default=None, action='store',
        choices=['config'],
        help="""detailed help topic.""")

    # parse the command
    args = vars(parser.parse_args())
    if args.get('func'):
        args['func'](args)
    elif args.get('version'):
        print('{} {}'.format(__package_name__, __version__))
    else:
        parser.parse_args(['-h'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
from setuptools import setup, find_packages
from inspect import cleandoc
import webscrapbook

long_description = open("README.md", encoding="utf-8").read()

setup(
    name=webscrapbook.__package_name__,
    version=webscrapbook.__version__,
    description=cleandoc(webscrapbook.__doc__),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=webscrapbook.__author__,
    author_email=webscrapbook.__author_email__,
    url=webscrapbook.__homepage__,
    license=webscrapbook.__license__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Natural Language :: English",
        "Topic :: Database",
        "Topic :: Internet",
        ],
    python_requires='~=3.5',
    install_requires=[
        "lxml >= 4.0",
        "bottle >= 0.12",
        'commonmark >= 0.8',
        ],
    packages=find_packages(),
    package_data={
        'webscrapbook': [
            'resources/*.*',
            'themes/default/static/*.*',
            'themes/default/templates/*.*',
            ],
        },
    entry_points={
        'console_scripts': [
            'webscrapbook = webscrapbook.cli:main',
            'wsb = webscrapbook.cli:main',
            ],
        },
    )



#!/usr/bin/env python3
"""Miscellaneous utilities
"""
import sys, os
import subprocess
from collections import namedtuple
from html.parser import HTMLParser
from lxml import etree
import zipfile
import math
import re
import hashlib
import time
from secrets import token_urlsafe
from urllib.parse import quote, unquote


#########################################################################
# URL and string
#########################################################################

def get_breadcrumbs(path, base='', topname='.', subarchivepath=None):
    """Generate (label, subpath, sep, is_last) tuples.
    """
    base = base.rstrip('/') + '/'
    subpathfull = path.strip('/')

    if subarchivepath is None:
        # /path/to/directory/
        archivepath = None
    elif subarchivepath == "":
        # /path/to/archive.ext!/
        archivepath = subpathfull
    else:
        # /path/to/archive.ext!/subarchivepath/
        archivepath = subpathfull[0:-(len(subarchivepath) + 1)]

    if subpathfull:
        yield (topname, base, '/', False)
        subpaths = []
        parts = subpathfull.split('/');
        parts_len = len(parts)
        for idx, part in enumerate(parts):
            subpaths.append(part)
            subpath = '/'.join(subpaths)
            if subpath == archivepath:
                yield (part[:-1], base + subpath + '/', '!/', idx == parts_len - 1)
            else:
                yield (part, base + subpath + '/', '/', idx == parts_len - 1)
    else:
        yield (topname, base, '/', True)


#########################################################################
# Filesystem related manipulation
#########################################################################

FileInfo = namedtuple('FileInfo', ['name', 'type', 'size', 'last_modified'])


def launch(path):
    """Launch a file or open a directory in the explorer.
    """
    if sys.platform == "win32":
        os.startfile(path)
    elif sys.platform == "darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])


def view_in_explorer(path):
    """Open the parent directory of a file or directory
       in the explorer.
    """
    if sys.platform == "win32":
        subprocess.Popen(["explorer", "/select,", path])
    elif sys.platform == "darwin":
        try:
            subprocess.Popen(["open", "-R", path])
        except:
            # fallback for older OS X
            launch(os.path.dirname(path))
    else:
        try:
            subprocess.Popen(["nautilus", "--select", path])
        except:
            # fallback if no nautilus
            launch(os.path.dirname(path))


def file_info(file):
    """Read basic file information.
    """
    name = os.path.basename(file)

    if not os.path.lexists(file):
        type = None
    elif os.path.islink(file):
        type = 'link'
    elif os.path.isdir(file):
        type = 'dir'
    elif os.path.isfile(file):
        type = 'file'
    else:
        type = 'unknown'

    try:
        statinfo = os.stat(file)
        size = statinfo.st_size if type is 'file' else None
        last_modified = statinfo.st_mtime
    except:
        # unexpected error when getting stat info
        size = None
        last_modified = None

    return FileInfo(name=name, type=type, size=size, last_modified=last_modified)


def listdir(root):
    """Generates FileInfo(s) and omit invalid entries.
    """
    for filename in os.listdir(root):
        file = os.path.join(root, filename)
        info = file_info(file)
        if info.type is None: continue
        yield info


def format_filesize(bytes, si=False):
    """Convert file size from bytes to human readable presentation.
    """
    try:
        bytes = int(bytes)
    except:
        return ''

    if si:
        thresh = 1000
        units = ['B', 'kB','MB','GB','TB','PB','EB','ZB','YB']
    else:
        thresh = 1024
        units =  ['B', 'KB','MB','GB','TB','PB','EB','ZB','YB']

    e = math.floor(math.log(bytes) / math.log(thresh))
    n = bytes / math.pow(thresh, e)
    tpl = '{:.1f} {}' if (e >=1 and n < 10) else '{:.0f} {}'
    return tpl.format(n, units[e])


#########################################################################
# ZIP handling
#########################################################################

class ZipDirNotFoundError(Exception):
    pass


def zip_file_info(zip, subpath, check_missing_dir=False):
    """Read basic file information from ZIP.

    subpath: 'dir' and 'dir/' are both supported
    """
    if not isinstance(zip, zipfile.ZipFile):
        zip = zipfile.ZipFile(zip)

    subpath = subpath.rstrip('/')
    basename = os.path.basename(subpath)

    try:
        info = zip.getinfo(subpath)
        lm = info.date_time
        epoch = int(time.mktime((lm[0], lm[1], lm[2], lm[3], lm[4], lm[5], 0, 0, -1)))
        return FileInfo(name=basename, type='file', size=info.file_size, last_modified=epoch)
    except KeyError:
        pass

    try:
        info = zip.getinfo(subpath + '/')
        lm = info.date_time
        epoch = int(time.mktime((lm[0], lm[1], lm[2], lm[3], lm[4], lm[5], 0, 0, -1)))
        return FileInfo(name=basename, type='dir', size=None, last_modified=epoch)
    except KeyError:
        pass

    if check_missing_dir:
        base = subpath + '/'
        for entry in zip.namelist():
            if entry.startswith(base):
                return FileInfo(name=basename, type='dir', size=None, last_modified=None)

    return FileInfo(name=basename, type=None, size=None, last_modified=None)


def zip_listdir(zip, subpath):
    """Generates FileInfo(s) and omit invalid entries.

    Raise ZipDirNotFoundError if subpath does not exist. 

    NOTE: It is possible that entry mydir/ does not exist while
    mydir/foo.bar exists. Check for matching subentries fo make sure whether
    the directory exists.
    """
    if not isinstance(zip, zipfile.ZipFile):
        zip = zipfile.ZipFile(zip)

    base = subpath.rstrip('/')
    if base: base += '/'
    base_len = len(base)
    dir_exist = not base
    entries = {}
    for filename in zip.namelist():
        if not filename.startswith(base):
            continue

        if filename == base:
            dir_exist = True
            continue

        subpath = filename[base_len:]
        entry, _, _ = subpath.partition('/')
        entries.setdefault(entry, True)

    if not len(entries) and not dir_exist:
        raise ZipDirNotFoundError('Directory "{}/" does not exist in the zip.'.format(base))

    for entry in entries:
        info = zip_file_info(zip, base + entry)

        if info.type is None:
            yield FileInfo(name=entry, type='dir', size=None, last_modified=None)
        else:
            yield info


#########################################################################
# HTML manipulation
#########################################################################

MetaRefreshInfo = namedtuple('MetaRefreshInfo', ['time', 'target'])


class MetaRefreshParser(HTMLParser):
    """Retrieve meta refresh target from HTML.
    """
    def __init__(self):
        self.meta_refresh_stack = []
        super().__init__()

    def handle_starttag(self, tag, attrs):
        if tag != 'meta':
            return

        attrs = dict(attrs)

        if attrs.get('http-equiv', '').lower().strip() != 'refresh':
            return

        time, _, content = attrs.get('content', '').partition(';')

        try:
            time = int(time)
        except ValueError:
            time = 0

        m = re.match(r'^\s*url\s*=\s*(.*?)\s*$', content)
        target = m.group(1) if m else None
        self.meta_refresh_stack.append(MetaRefreshInfo(time=time, target=target))


def parse_meta_refresh(fh):
    """Retrieve meta refresh target from a file.

    fh: a file or file handler.
    """
    if type(fh) is str:
        try:
            fh = open(fh, 'rb')
        except:
            fh = None

    if fh is not None:
        try:
            parser = MetaRefreshParser()
            parser.feed(fh.read().decode('UTF-8'))
            for mr in parser.meta_refresh_stack:
                if mr.time == 0 and mr.target is not None:
                    return mr
        except:
            pass

    return MetaRefreshInfo(time=None, target=None)


#########################################################################
# MAFF manipulation
#########################################################################

MaffPageInfo = namedtuple('MaffPageInfo', ['title', 'originalurl', 'archivetime', 'indexfilename', 'charset'])

def get_maff_pages(file):
    """Get a list of pages (MaffPageInfo).
    """
    pages = []
    with zipfile.ZipFile(file) as zip:
        # get top folders and their content files
        topdirs = {}
        for entry in zip.namelist():
            topdir, sep, p = entry.partition('/')
            topdir = topdirs.setdefault(topdir + sep, [])
            if p: topdir.append(entry)

        # get index files
        for topdir in topdirs:
            rdf = topdir + 'index.rdf'
            try:
                with zip.open(rdf, 'r') as f:
                    meta = parse_maff_index_rdf(f)
                    f.close()

                if meta.indexfilename is not None:
                    pages.append(MaffPageInfo(
                            meta.title,
                            meta.originalurl,
                            meta.archivetime,
                            topdir + meta.indexfilename,
                            meta.charset,
                            ))
                    continue
            except:
                pass

            for entry in topdirs[topdir]:
                if entry.startswith(topdir + 'index.') and entry != topdir + 'index.rdf':
                    pages.append(MaffPageInfo(
                            None,
                            None,
                            None,
                            entry,
                            None,
                            ))

    return pages


def parse_maff_index_rdf(fh):
    """Read MAFF metadata from the given RDF file handler.
    """
    def load_attr(attr):
        try:
            node = root.find('./RDF:Description/MAF:' + attr, ns)
            return node.attrib['{' + ns['RDF'] + '}' + 'resource']
        except:
            raise
            return None

    ns = {
        'MAF': "http://maf.mozdev.org/metadata/rdf#",
        'NC': "http://home.netscape.com/NC-rdf#",
        'RDF': "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        }

    text = fh.read().decode('UTF-8')
    root = etree.XML(text)

    return MaffPageInfo(
            load_attr('title'),
            load_attr('originalurl'),
            load_attr('archivetime'),
            load_attr('indexfilename'),
            load_attr('charset'),
            )


#########################################################################
# Encrypt and security
#########################################################################

class Encrypt():
    """Simple hash encryption with salt.
    """
    def md5(self, text, salt=''):
        return hashlib.md5((text + salt).encode('UTF-8')).hexdigest()

    def sha1(self, text, salt=''):
        return hashlib.sha1((text + salt).encode('UTF-8')).hexdigest()

    def sha224(self, text, salt=''):
        return hashlib.sha224((text + salt).encode('UTF-8')).hexdigest()

    def sha256(self, text, salt=''):
        return hashlib.sha256((text + salt).encode('UTF-8')).hexdigest()

    def sha384(self, text, salt=''):
        return hashlib.sha384((text + salt).encode('UTF-8')).hexdigest()

    def sha512(self, text, salt=''):
        return hashlib.sha512((text + salt).encode('UTF-8')).hexdigest()

    def sha3_224(self, text, salt=''):
        return hashlib.sha3_224((text + salt).encode('UTF-8')).hexdigest()

    def sha3_256(self, text, salt=''):
        return hashlib.sha3_256((text + salt).encode('UTF-8')).hexdigest()

    def sha3_384(self, text, salt=''):
        return hashlib.sha3_384((text + salt).encode('UTF-8')).hexdigest()

    def sha3_512(self, text, salt=''):
        return hashlib.sha3_512((text + salt).encode('UTF-8')).hexdigest()

    def plain(self, text, salt=''):
        return text + salt

    def encrypt(self, text, salt='', method='plain'):
        fn = getattr(self, method)

        if not callable(fn):
            print('Encrypt method "{}" not implemented, fallback to "plain".'.format(method), file=sys.stderr)
            fn = self.plain

        return fn(text, salt)

encrypt = Encrypt().encrypt


class TokenHandler():
    """Handle security token validation to avoid XSRF attack.
    """
    def __init__(self, cache_dir):
        self.cache_dir = cache_dir
        self.last_purge = 0

    def acquire(self, now=None):
        if now is None:
            now = int(time.time())

        self.check_delete_expire(now)

        token = token_urlsafe()
        token_file = os.path.join(self.cache_dir, token)
        while os.path.lexists(token_file):
            token = token_urlsafe()
            token_file = os.path.join(self.cache_dir, token)

        os.makedirs(os.path.dirname(token_file), exist_ok=True)
        open(token_file, 'w', encoding='UTF-8').write(str(now + self.DEFAULT_EXPIRY))

        return token

    def validate(self, token, now=None):
        if now is None:
            now = int(time.time())

        token_file = os.path.join(self.cache_dir, token)

        try:
            expire = int(open(token_file, 'r', encoding='UTF-8').read())
        except FileNotFoundError:
            return False

        if now >= expire:
            self.delete(token)
            return False

        return True

    def delete(self, token):
        token_file = os.path.join(self.cache_dir, token)

        try:
            os.remove(token_file)
        except:
            pass

    def delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        try:
            for token_file in os.listdir(self.cache_dir):
                token_file = os.path.join(self.cache_dir, token_file)
                try:
                    expire = int(open(token_file, 'r', encoding='UTF-8').read())
                except:
                    continue

                if now >= expire:
                    os.remove(token_file)
        except FileNotFoundError:
            pass

    def check_delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        if now >= self.last_purge + self.PURGE_INTERVAL:
            self.last_purge = now
            self.delete_expire(now)

    PURGE_INTERVAL = 3600  # in seconds
    DEFAULT_EXPIRY = 1800  # in seconds



#!/usr/bin/env python3
"""Server backend of WebScrapBook toolkit.
"""
import os
import time
import webbrowser
from threading import Thread

# dependency
import bottle

# this package
from . import *
from .app import init_app


class WSBServer(bottle.ServerAdapter):
    """Tweaked from bottle.WSGIRefServer

    - Remove starting message of bottle.
    - Support of SSL.
    - Support multi-threading.
    """
    def run(self, app): # pragma: no cover
        #! we have bypassed bottle
        self.quiet = False

        from wsgiref.handlers import BaseHandler
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        import multiprocessing.pool
        import socket
        import ssl

        class FixedHandler(WSGIRequestHandler):
            def address_string(self): # Prevent reverse DNS lookups please.
                return self.client_address[0]
            def log_request(*args, **kw):
                if not self.quiet:
                    return WSGIRequestHandler.log_request(*args, **kw)

        class ThreadPoolWSGIServer(WSGIServer):
            """WSGI-compliant HTTP server.  Dispatches requests to a pool of threads.
            
            ref: https://github.com/RonRothman/mtwsgi
            """

            def __init__(self, thread_count=None, *args, **kwargs):
                """If 'thread_count' == None, we'll use multiprocessing.cpu_count() threads."""
                WSGIServer.__init__(self, *args, **kwargs)
                self.thread_count = thread_count
                self.pool = multiprocessing.pool.ThreadPool(self.thread_count)

            # Inspired by SocketServer.ThreadingMixIn.
            def process_request_thread(self, request, client_address):
                try:
                    self.finish_request(request, client_address)
                    self.shutdown_request(request)
                except:
                    self.handle_error(request, client_address)
                    self.shutdown_request(request)

            def process_request(self, request, client_address):
                self.pool.apply_async(self.process_request_thread, args=(request, client_address))


        def make_server(host, port, app, server_cls, handler_class, thread_count=None):
            '''Create a new WSGI server listening on `host` and `port` for `app`'''
            httpd = server_cls(thread_count, (host, port), handler_class)
            httpd.set_app(app)
            return httpd

        BaseHandler.http_version = self.options.get('http_version', "1.1")
        handler_cls = self.options.get('handler_class', FixedHandler)
        server_cls  = self.options.get('server_class', ThreadPoolWSGIServer)

        if ':' in self.host: # Fix wsgiref for IPv6 addresses.
            if getattr(server_cls, 'address_family') == socket.AF_INET:
                class server_cls(server_cls):
                    address_family = socket.AF_INET6

        srv = make_server(self.host, self.port, app, server_cls, handler_cls, self.options.get('threads') or None)

        #! add SSL support here
        if self.options.get('ssl_on'):
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(
                certfile=self.options.get('ssl_cert'),
                keyfile=self.options.get('ssl_key'),
                password=self.options.get('ssl_pw'),
                )
            srv.socket = context.wrap_socket(srv.socket,
                    server_side=True,
                    )
            srv.base_environ['HTTPS'] = 'on'

        srv.serve_forever()


def launch_browser(url, browser_path, **kwargs):
    browser = webbrowser.get(browser_path)
    browser.open(url, **kwargs)


def serve(root, **kwargs):
    # switch to the specified directory and reload configs
    os.chdir(root)
    config.load()

    # set params
    host = config['server']['host'] or ''
    port = config['server'].getint('port')
    ssl_on = config['server'].getboolean('ssl_on')
    ssl_key = config['server']['ssl_key']
    ssl_cert = config['server']['ssl_cert']
    threads = config['server'].getint('threads')
    scheme = 'https' if ssl_on else 'http'

    # start server
    thread = Thread(target=init_app().run, kwargs={
        'host': host,
        'port': port,
        'server': WSBServer,
        'ssl_on': ssl_on,
        'ssl_key': ssl_key,
        'ssl_cert': ssl_cert,
        'threads': threads,
        'quiet': True,  # bypass Bottle init message, will be reset in server
        })
    thread.daemon = True
    thread.start()

    print('WebScrapBook server starting up...')
    print('Document Root: {}'.format(os.path.abspath(root)))
    print('Listening on {scheme}://{host}:{port}'.format(
            scheme=scheme, host=host, port=port))
    print('Hit Ctrl-C to shutdown.')
    print()

    # launch the browser
    if config['server'].getboolean('browse'):
        base = config['app']['base'].rstrip('/')
        index = config['browser']['index'].lstrip('/')
        path = base + (('/' + index) if index else '')

        url = '{scheme}://{host}{port}{path}'.format(
                scheme=scheme,
                host=host,
                port='' if (not ssl_on and port == 80) or (ssl_on and port == 443)
                        else ':' + str(port),
                path=path,
                )

        browser_path = config['browser']['command'] or None

        kwargs = {
            'new': config['browser'].getint('new'),
            'autoraise': config['browser'].getboolean('top'),
            }

        thread = Thread(target=launch_browser, args=[url, browser_path], kwargs=kwargs)
        thread.daemon = True
        thread.start()

        print('Launching browser at {url} ...'.format(url=url))

    try:
        while True: time.sleep(100)
    except (KeyboardInterrupt, SystemExit):
        print('')
        print('Keyboard interrupt received, shutting down server.')


def main():
    serve(".")


if __name__ == '__main__':
    main()
